//
// This fuzzer is an improvement of cross_fuzz_randomized_20100729_seed.html
// that makes it considerably easier to reproduce fault conditions,
// especially when running it from file:///. This is achieved by minimizing
// the risk of DOM crawl desynchronization by resyncing the PRNG when
// returning from a particular recursion level, and after shuffling arrays.
// 
// Important advice on setting up the fuzzer to get useful results:
// 
// 1) wget -r -np the entire http://lcamtuf.coredump.cx/cross_fuzz/
//    directory to a location in file:///
// 
// 2) In your test setup, set the home page to about:blank, disable safe
//    browsing, configure the browser to purge the cache on exit, disable
//    pop-up blocking, proxy autodetection, update checking, and eliminate
//    other factors that may be introducing nondeterministic latency.
//
//    For MSIE, you also need to do:
//
//    diff -u cross_fuzz_randomized_20100729_seed.html cross_fuzz_msie_randomized_seed.html
//
//    ...and apply the resulting patch to this file.
//
//    For MSIE and WebKit browsers, you also need to make a configuration
//    change to allow file:/// access from JavaScript; or - less preferably -
//    set up a local web server instead. Advice on how to do this is not
//    provided here.
// 
// 3) Start the fuzzer inside a browser in the "private" browsing mode if
//    possible (so that the influence of cache is minimized). Pick a random
//    32-bit seed for every fuzzing cycle.
//    
//    When done, navigate to the fuzzer via:
// 
//    file:///wherever_you_put_it/cross_fuzz_randomized_20110105_seed.html#1234
//  
//    ...where 1234 is your selected seed.
// 
// 4) Collect seed and crash signature pairs when repeating step #3. For each
//    crash, pick a seed that resulted in that crash in the shortest amount
//    of time. Use this seed to reproduce the fault for debugging. If it
//    does not reproduce, use the second shortest time, etc.
// 
// Happy fuzzing.
//

var FAN_LIMIT     = 8;    /* Object crawl fanout limit                 */
var MAX_LEVEL     = 5;    /* Maximum object nest level to crawl        */
var MAX_RET_LEVEL = 1;    /* Maximum ret val object level to crawl     */

var TWEAK_ODDS    = 2;    /* Property tweak probability when crawling  */
var CALL_ODDS     = 2;    /* Method call probability                   */

var REF_ODDS      = 5;    /* Object reference storing probability      */
var NONOBJ_ODDS   = 20;   /* Non-object reference storing probability  */
var INTER_ODDS    = 2;    /* Odds of using interesting_vals vs refs    */

var TRASH_ODDS    = 8;    /* Target window trashing probability        */
var RESET_ODDS    = 2;    /* Odds of respawning target windows         */

var PARAMS        = 4;    /* Number of params to make up for methods   */
var MAX_REFS      = 200;  /* Maximum number of refs to keep            */
var KEEP_REFS     = 100;  /* Number of refs to move across cycles      */

var Dump = (typeof dump == 'undefined') ? (function () {}) : dump;

/* Script replay stuff.  Send string to stdout which can be extracted
  to create the replay script.  The second parameter controls the
  indentation prior to the output and after.  + means increase
  indentation, - means decrease indentation. The first character
  controls the prior indentiation while the second controls the after.
  Interesting replay scripts cause crashes and to help recover the
  recursive nature of the replay script, we annotate the open and closing
  blocks to help identify the missing closing function calls.
 */

var replay_indent = '';
var replay_block_stack = [];
var replay_block_id = 0;

function replay(string, action) {
  if (!string)
    return;

  action = action || "";

  var spaces = "  ";

  if (action) {
    switch(action.substring(0,1)) {
    case "+":
      replay_indent += spaces;
      Dump("cross_fuzz_replay:" + replay_indent + "/* open block " + replay_block_id + " */\n");
      replay_block_stack.push(replay_block_id++);
      break
    case "-":
      Dump("cross_fuzz_replay:" + replay_indent + "/* close block " + replay_block_stack.pop() + " */\n");
      replay_indent = replay_indent.slice(0, -spaces.length);
      break;
    }
  }

  Dump("cross_fuzz_replay:" + replay_indent + string + '\n');

  if (action) {
    switch(action.substring(1,2)) {
    case "+":
      replay_indent += spaces;
      Dump("cross_fuzz_replay:" + replay_indent + "/* open block " + replay_block_id + " */\n");
      replay_block_stack.push(replay_block_id++);
      break
    case "-":
      Dump("cross_fuzz_replay:" + replay_indent + "/* close block " + replay_block_stack.pop() + " */\n");
      replay_indent = replay_indent.slice(0, -spaces.length);
      break;
    }
  }

}

/**************************
 * Various crawl settings *
 **************************/

replay("var interesting_vals = [", " +");
replay("0, 1, 1e6, -1e6, 1e-6, 1e100, null, undefined, 'pink', screen, Infinity, false, true, eval, [], {},");
replay("4500000000, 2200000000, -2200000000, -4500000000");
replay("];", "- ");

var interesting_vals = [
  0,             //  0
  1,             //  1
  1e6,           //  2
 -1e6,           //  3
  1e-6,          //  4
  1e100,         //  5
  null,          //  6
  undefined,     //  7
  'pink',        //  8
  screen,        //  9
  Infinity,      // 10
  false,         // 11
  true,          // 12
  eval,          // 13
  [],            // 14
  {},            // 15
  4500000000,    // 16
  2200000000,    // 17
 -2200000000,    // 18
 -4500000000     // 19
];

var object_blacklist = { 

  /* Crashes */
  'resizeTo' : 0,               // bug 622165
  'resizeBy' : 0,               // bug 622165
  'importUserCertificates' : 0, // bug 624493
  'setTimeout' : 0,             // bug 622593

  /* Problems? */
  'files' : 0,  // file dialog on linux blocks

  /* Properties */

  'fuzz1_visited' : 0,
  'fuzz2_visited' : 0,
  'fuzz3_visited' : 0,
  'ref_visited' : 0,

  'location' : 0,
  'name' : 0,
  'opener' : 0,
  'URL' : 0,
  'onbeforeunload' : 0,
  'onunload': 0,

  'innerHTML' : 0,
  'outerHTML' : 0,
  'innerText' : 0,
  'textContent' : 0,

  'Components': 0,
  'controllers' : 0,

  'lastChild' : 0,
  'previousSibling': 0,
  'previousElementSibling': 0,
  'parentNode' : 0,
  'parentTextEdit' : 0,
  'parentElement' : 0,
  'ownerDocument' : 0,
  'document' : 1,

  /* Methods */

  'cloneNode' : 0,
  'open' : 0,
  'close' : 0,
  'print' : 0,
  'alert': 0,
  'prompt' : 0,
  'showModalDialog' : 0,
  'confirm' : 0

};

/*****************************************
 * Generate random integer (0 ... mod-1) *
 *****************************************/

function R(mod) {
  return genrand_int32() % mod;
}

/**************************
 * Eliminate all random   *
 **************************/
function Random() {
  return R(1000000000)/1000000000
}

/*************
 * Log stuff *
 *************/

var log_box;
var message_data = '';

function LOG(message) {

  try {
    Dump('cross_fuzz: ' + message + '\n');
  }
  catch(e) {
    Dump('cross_fuzz: exception printing message.\n');
  }

  /* TODO: Find a way to log stuff. */

return;

  message_data += '* ' + message + '\n';

  log_box.value = message_data.substr(message_data.length - 2048);
  log_box.scrollTop = log_box.scrollHeight;

}

/*************
 * GC helper *
 *************/

var gc_string;

function toggle_gc() {

  for (var i = 0; i < 10000; i++) {
    var s = new String("AAAA" + Random());
    eval("gc_string = s");
  }

}

/******************
 * Array shuffler *
 ******************/

function shuffle_array(arr) {
  var cnt = arr.length;
  var ret_seed = genrand_int32();

  while (cnt--) {
    var src = R(arr.length), dst = R(arr.length);
    var tmp = arr[src];

    arr[src] = arr[dst];
    arr[dst] = tmp;
  }

  init_genrand(ret_seed);

}


/*************************
 * Reopen target windows *
 *************************/

replay("var t1, t2;");
var t1, t2;

function new_targets() {

  LOG('Respawning targets...');

  replay("(function () {  // args: none", " +");
  replay("try {", " +");
  replay("t1.close();");
  replay("t2.close();");
  replay("} catch (e) { }", "- ");
  replay("try {", " +");
  replay("t1.close();");
  replay("t2.close();");
  replay("} catch (e) { }", "- ");

  /* TODO: Add more interesting document types. */

  switch (R(4)) {

    case 0:
      if ('v' != '\v') {
        replay("t1 = window.open('targets/target.svg', 't1');");
        t1 = window.open('targets/target.svg', 't1');
        LOG('Target 1: SVG');
      } else {
        replay("t1 = window.open('targets/target_strict2.html', 't1');");
        t1 = window.open('targets/target_strict2.html', 't1');
        LOG('Target 1: HTML strict');
      }
      break;

    default:
      replay("t1 = window.open('targets/target2.html', 't1');");
      t1 = window.open('targets/target2.html', 't1');
      LOG('Target 1: HTML');
  
  }

  switch (R(4)) {

    case 0:
      if ('v' != '\v') {
        replay("t2 = window.open('targets/target.svg', 't2');");
        t2 = window.open('targets/target.svg', 't2');
        LOG('Target 2: SVG');
      } else {
        dump_relay("t2 = window.open('targets/target_strict2.html', 't2');");
        t2 = window.open('targets/target_strict2.html', 't2');
        LOG('Target 2: HTML strict');
      }
      break;

    default:
      replay("t2 = window.open('targets/target2.html', 't2');");
      t2 = window.open('targets/target2.html', 't2');
      LOG('Target 2: HTML');
  
  }

  replay("try { t1.opener = null; } catch (e) { }");
  replay("try { t2.opener = null; } catch (e) { }");
  replay("})();", "- ");

  try { t1.opener = null; } catch (e) { }
  try { t2.opener = null; } catch (e) { }

}

/*******************************************************
 * Trash target window in one of several possible ways *
 *******************************************************/

function trash_target(target) {

  replay("(function () { // args: none", " +");
  replay("try {", " +");

  try {

    switch (R(3)) {

      case 0: 
        LOG('Clobbering target with document.write()');
        replay(target.name + ".document.write('Hi mom!');");
        target.document.write('Hi mom!');
        break;

      case 1:
        LOG('Clobbering target with document.body.innerHTML');
        replay(target.name + ".document.body.innerHTML = 'Hi mom';");
        target.document.body.innerHTML = 'Hi mom'; 
        break;

      case 2:
        LOG('Clobbering target with window.close()');
        replay(target.name + ".close();");
        target.close();
        break;

    }

  } catch (e) { }

  replay("} catch(e) { }", "- ");
  replay("})();", "- ");
}

/*****************************
 * Maybe add a new reference *
 *****************************/

function maybe_add_ref(obj, add_set) {

  if (R(REF_ODDS) != 0) return;

  /* Be more conservative about adding non-objects. */
  if (typeof obj != 'object' && R(NONOBJ_ODDS) != 0) return;

  replay("try {", " +");

  try {

    try {
      if (obj.ref_visited) {
        replay("} catch(e) { }", "- ");
        return;
      }
      obj.ref_visited = 1;
    } catch (e) { }

    LOG('+++ Adding reference ' + obj + ' (' + add_set.length + ') +++');

    var r = R(MAX_REFS);

    if (add_set.length > MAX_REFS) {
      replay("add_set[" + r + "] = obj;");
      add_set[r] = obj; 
    } else {
      replay("add_set.push(obj);");
      add_set.push(obj);
    }

  } catch (e) { }

  replay("} catch(e) { }", "- ");
}

/*****************************
 * Crawl, collect properties *
 *****************************/

var crawl_history = [];
var cur_id;

function crawl_properties(path, target, level, add_set) {

  var members = [];
  var cur_fan = 0;

  var ret_seed = genrand_int32();

  LOG('-- PROPERTY CRAWL (' + level + '): ' + path + ' --');

  if (level > MAX_LEVEL) {
    LOG("*** NESTING LEVEL EXCEEDED ***");
    return;
  }

  try {
    for (var name in target) members.push(name);
  } catch (e) {
    LOG("Boo - cannot enumerate: " + e);
    return;
  }

  /* Make sure this is not a duplicate. */

  if (level == 0) cur_id = Random();

  try {
    if (target.fuzz1_visited == cur_id) {
      LOG('*** Already crawled (1) ***');
      return;
    }
    target.fuzz1_visited = cur_id;
  } catch (e) { }

  shuffle_array(members);

  replay("var cur_value;");
  replay("var name;");

  for (var num in members) {
    var name = members[num];
    var cur_value = null;

    if (name == '0' || 
        (object_blacklist[name] != undefined && level >= object_blacklist[name])) {
      LOG('Skipping: ' + path + '.' + name);
      continue;
    }

    LOG('Trying: ' + path + '.' + name);

    // must emit replay code first to make sure it is emitted.
    replay("cur_value = null;");
    replay("name = '" + name + "';");
    replay("try {", " +");
    replay("cur_value = eval('target.' + name);");
    replay("} catch (e) {", "-+");
    //replay("Dump('...received exception (' + e + ')\\n');");
    replay("}", "- ");

    try {
      cur_value = eval('target.' + name);
      //LOG('...result is "' + cur_value + '" (' + typeof cur_value + ')');
    } catch (e) {
      //LOG('...received exception (' + e + ')');
    }

    if (cur_value != null) {

      if (typeof cur_value != 'function') {
        replay("(function (obj, add_set) { // args: cur_value, add_set", " +");
        maybe_add_ref(cur_value, add_set);
        replay("})(cur_value, add_set);", "- ");
      }

      /* Recurse into objects */

      if (typeof cur_value == 'object' && cur_fan < FAN_LIMIT) {

        cur_fan++;
        replay("(function (path, target, add_set) { // args: path + '.' + name, cur_value, add_set", " +");
        crawl_properties(path + '.' + name, cur_value, level + 1, add_set);
        replay("})(path + '.' + name, cur_value, add_set);", "- ");
        LOG('-- BACK TO PROPERTY CRAWL (' + level + '): ' + path + ' --');

      }

    }

  }

  init_genrand(ret_seed);

}

/********************
 * Tweak properties *
 ********************/

function tweak_properties(path, target, level, use_set) {

  var members = [];
  var cur_fan = 0;

  var ret_seed = genrand_int32();

  LOG('-- PROPERTY TWEAK (' + level + '): ' + path + ' --');

  if (level > MAX_LEVEL) {
    LOG("*** NESTING LEVEL EXCEEDED ***");
    return;
  }

  try {
    for (var name in target) members.push(name);
  } catch (e) {
    LOG("Boo - cannot enumerate: " + e);
    return;
  }

  /* Make sure this is not a duplicate. */

  if (level == 0) cur_id = Random();

  try {
    if (target.fuzz2_visited == cur_id) {
      LOG('*** Already crawled (1) ***');
      return;
    }
    target.fuzz2_visited = cur_id;
  } catch (e) { }

  shuffle_array(members);

  replay("var orig_value;");
  replay("var name;");

  for (var num in members) {

    var name = members[num];
    var orig_value = null;

    if (name == '0' ||
        (object_blacklist[name] != undefined && level >= object_blacklist[name])) {
      LOG('Skipping: ' + path + '.' + name);
      continue;
    }

    // must emit replay code first to make sure it is emitted.
    replay("orig_value = null;");
    replay("name = '" + name + "';");
    replay("try {", " +");
    replay("orig_value = eval('target.' + name);");
    replay("} catch (e) {", "-+");
    //replay("Dump('...received exception (' + e + ')\\n');");
    replay("}", "- ");

    try {
      orig_value = eval('target.' + name);
    } catch (e) { 
      //LOG('...received exception (' + e + ')');
      continue; 
    }

    /* Leave functions alone!!! */
    if (typeof orig_value == 'function') continue;

    LOG('Trying: ' + path + '.' + name);

    replay("try {", " +");

    if (R(TWEAK_ODDS) == 0) try {

      var r;

      if (use_set.length != 0 && R(INTER_ODDS) != 0) {
        r = R(use_set.length)
        replay("eval('target.' + name + ' = use_set[" + r + "]');");
        eval('target.' + name + ' = use_set[' + r + ']');
      } else {
        r = R(interesting_vals.length);
        replay("eval('target.' + name + ' = interesting_vals[" + r + "]');");
        eval('target.' + name + ' = interesting_vals[' + R(interesting_vals.length) + ']');
      }

    } catch (e) {
      //LOG('...received exception (' + e + ')');
    }

    replay("} catch(e) {", "-+");
    //replay("Dump('...received exception (' + e + ')\\n');");
    replay("}", "- ");

    /* Recurse into objects. */

    if (orig_value != null && typeof orig_value == 'object' && cur_fan < FAN_LIMIT) {

      cur_fan++;
      replay("(function (path, target, use_set) { // args: path + '.' + name, orig_value, use_set", " +");
      tweak_properties(path + '.' + name, orig_value, level + 1, use_set);
      replay("})(path + '.' + name, orig_value, use_set);", "- ")
      LOG('-- BACK TO PROPERTY TWEAK (' + level + '): ' + path + ' --');

    }

  }

  init_genrand(ret_seed);

}


/***********************
 * Crawl, call methods *
 ***********************/

function call_methods(path, target, level, ret_level, use_set, add_set) {

  var members = [];
  var cur_fan = 0;

  var ret_seed = genrand_int32();

  LOG('-- METHOD CRAWL (' + level + '): ' + path + ' --');

  if (level > MAX_LEVEL) {
    LOG("*** NESTING LEVEL EXCEEDED ***");
    return;
  }

  if (ret_level > MAX_RET_LEVEL) {
    LOG("*** RET NESTING LEVEL EXCEEDED ***");
    return;
  }

  try {
    for (var name in target) members.push(name);
  } catch (e) {
    LOG("Boo - cannot enumerate: " + e);
    return;
  }

  /* Make sure this is not a duplicate. */

  if (level + ret_level == 0) cur_id = Random();

  try {
    if (target.fuzz3_visited == cur_id) {
      LOG('*** Already crawled (1) ***');
      return;
    }
    target.fuzz3_visited = cur_id;
  } catch (e) { }

  shuffle_array(members);

  replay("var cur_value;");
  replay("var name;");

  for (var num in members) {

    var name = members[num];
    var cur_value = null;

    if (name == '0' ||
        (object_blacklist[name] != undefined && level >= object_blacklist[name])) {
      LOG('Skipping: ' + path + '.' + name);
      continue;
    }

    // must emit replay code first to make sure it is emitted.
    replay("cur_value = null;");
    replay("name = '" + name + "';");
    replay("try {", " +");
    replay("cur_value = eval('target.' + name);");
    replay("} catch (e) {", "-+");
    //replay("Dump('...received exception (' + e + ')\\n');");
    replay("}", "- ");

    try {
      cur_value = eval('target.' + name);
    } catch (e) { 
      //LOG('...received exception (' + e + ')');
      continue; 
    }

    if (cur_value == null) continue;

    if (R(CALL_ODDS) == 0 && typeof cur_value == 'function') {

      var ret_value = null;
      var par_str = "";

      /* Construct parameters for the call */

      for (var i = 0; i < PARAMS; i++) {

        if (use_set.length != 0 && R(INTER_ODDS) != 0)
          par_str += 'use_set[' + R(use_set.length) + ']';
        else
          par_str += 'interesting_vals[' + R(interesting_vals.length) + ']';

        if (i + 1 != PARAMS) par_str += ", ";

      }

      LOG('Trying: ' + path + '.' + name);

      // must emit replay code first to make sure it is emitted.
      // XXX: ?
      replay("ret_value = null;");
      replay("name = '" + name + "';");
      replay("try {", " +");
      replay("ret_value = eval('target.' + name + '(" + par_str + ")');")
      replay("} catch (e) {", "-+");
      //replay("Dump('...received exception (' + e + ')\\n');");
      replay("}", "- ");

      try {

        ret_value = eval('target.' + name + '(' + par_str + ')');
        //LOG('...result is "' + ret_value + '" (' + typeof ret_value + ')');

      } catch (e) {
        //LOG('...received exception (' + e + ')');
      }

      if (ret_value != null) {

        if (typeof ret_value != 'function') {
          replay("(function (obj, add_set) { // args: ret_value, add_set", " +");
          maybe_add_ref(ret_value, add_set);
          replay("})(ret_value, add_set);", "- ");
        }

        /* Recurse into returned objects. */
 
        if (typeof ret_value == 'object' && cur_fan < FAN_LIMIT) {

          cur_fan++;
          replay("(function (path, target, use_set, add_set) { // args: path + '.[ret:' + name + ']' , ret_value, use_set, add_set", " +");
          call_methods(path + '.[ret:' + name + ']' , ret_value, level, ret_level + 1, use_set, add_set);
          replay("})(path + '.[ret:' + name + ']' , ret_value, use_set, add_set);", "- ");
          replay("(function (path, target, use_set) { // args: path + '.[ret:' + name + ']' , ret_value, use_set", " +");
          tweak_properties(path + '.[ret:' + name + ']' , ret_value, level + ret_level + 1, use_set); 
          replay("})(path + '.[ret:' + name + ']' , ret_value, use_set);", "- ");
          LOG('-- BACK TO METHOD CRAWL (' + level + '): ' + path + ' --');

        }

      }

    }

    /* Recurse into crawled objects. */

    if (typeof cur_value == 'object' && cur_fan < FAN_LIMIT) {

      cur_fan++;
      replay("(function (path, target, use_set, add_set) { // args: path + '.' + name, cur_value, use_set, add_set", " +");
      call_methods(path + '.' + name, cur_value, level + 1, ret_level, use_set, add_set);
      replay("})(path + '.' + name, cur_value, use_set, add_set);", "- ");
      LOG('-- BACK TO METHOD CRAWL (' + level + '): ' + path + ' --');

    }

  }

  init_genrand(ret_seed);

}

/*******************
 * Main event loop *
 *******************/

var state = 0;

replay("var cur_set = [];");
replay("var new_set = [];");

var cur_set = [];
var new_set = [];

var wait_cycles = 0;

function event_loop() {

  var ret_seed = genrand_int32();

  if (t1 == null) {
    LOG('Disable pop-up blocking.');
    setTimeout("quit()", 10);
    return;
  }

  if (wait_cycles > 0) {
    wait_cycles--;
    setTimeout('try { event_loop() } catch (e) { setTimeout("quit()", 10); LOG("FATAL_ERROR: " + e);}', 10);
    return;
  }

  replay("(function () { // args: none", " +");

  switch (state) {

    case 0:

      replay("(function (path, target, add_set) { // args: '[target1]', t1, cur_set", " +");
      crawl_properties('[target1]', t1, 0, cur_set);
      replay("})('[target1]', t1, cur_set);", "- ");
      break;

    case 1:
      replay("(function (path, target, use_set, add_set) { // args: '[target1]', t1, cur_set, cur_set", " +");
      call_methods('[target1]', t1, 0, 0, cur_set, cur_set);
      replay("})('[target1]', t1, cur_set, cur_set);", "- ");
      break;

    case 2:
      replay("(function (path, target, use_set) { // args: '[target1]', t1, cur_set", " +");
      tweak_properties('[target1]', t1, 0, cur_set);
      replay("})('[target1]', t1, cur_set);", "- ");
      break;

    case 3:
      replay("(function (path, target, use_set, add_set) { // args: '[target1]', t1, cur_set, cur_set", " +");
      call_methods('[target1]', t1, 0, 0, cur_set, cur_set);
      replay("})('[target1]', t1, cur_set, cur_set);", "- ");
      break;

    case 4:

      if (R(TRASH_ODDS) == 0) {
        trash_target(t1);
      }

      replay("toggle_gc();");
      toggle_gc();
      break;

    case 5:
      replay("(function (path, target, add_set) { // args: '[target2]', t2, new_set", " +");
      crawl_properties('[target2]', t2, 0, new_set);
      replay("})('[target2]', t2, new_set);", "- ");
      break;

    case 6:
      replay("(function (path, target, use_set, add_set) { // args: '[target2]', t2, cur_set, new_set", " +");
      call_methods('[target2]', t2, 0, 0, cur_set, new_set);
      replay("})('[target2]', t2, cur_set, new_set);", "- ");
      break;

    case 7:
      replay("(function (path, target, use_set) { // args: '[target2]', t2, new_set", " +");
      tweak_properties('[target2]', t2, 0, new_set);
      replay("})('[target2]', t2, new_set);", "- ");
      break;

    case 8:
      replay("(function (path, target, use_set, add_set) { // args: '[target2]', t2, cur_set, new_set", " +");
      call_methods('[target2]', t2, 0, 0, cur_set, new_set);
      replay("})('[target2]', t2, cur_set, new_set);", "- ");
      break;

    case 9:

      stat_box.innerHTML = cur_set.length;

      replay("cur_set = new_set.slice(-KEEP_REFS);");
      replay("new_set = [];");

      cur_set = new_set.slice(-KEEP_REFS);
      new_set = [];

      LOG('Kept ' + cur_set.length + ' references.');

      if (R(RESET_ODDS) == 0) {
        new_targets();
        replay("toggle_gc();");
        toggle_gc();
        wait_cycles = 50;
      }

      break;

  }
 
  state = (state + 1) % 10;

  replay("//### State " + state + " ###");
  LOG('### State ' + state + ' ###');

  setTimeout('try { event_loop() } catch (e) { setTimeout("quit()", 10); LOG("FATAL_ERROR: " + e);}', 10);

  replay("})();", "- ");

  init_genrand(ret_seed);

}

/***********************
 * Initialization code *
 ***********************/

function run_tests(run) {

  // always start with a standard window size
  if ('sizeToContent' in window)
    window.sizeToContent();

  if (!run && location.search.indexOf('run') == -1)
    return;

  if (typeof seed != "undefined")
    seed = seed;
  else if(location.hash) {
    seed = location.hash.substr(1);
  } else {
    seed = (new Date()).getTime() & 0xFFFFFFFF;
  }

  replay("var seed = " + seed + ';');

  LOG('seed = ' + seed);

  init_genrand(seed);
  location.hash = '#' + seed;

  replay("init_genrand(" + seed + ");");
  replay("location.hash = '#' + " + seed + ";");

  log_box  = document.getElementById('results');
  stat_box = document.getElementById('stats');

  new_targets();

  setTimeout('try { event_loop() } catch (e) { setTimeout("quit()", 10); LOG("FATAL_ERROR: " + e);}', 1000);

  // Terminate after 1 minutes. Any longer and the replay scripts
  // are too long.
  // XXX: set via url?
  setTimeout("quit()", 1000*60);

}

/****************************************
 * quit() uses Jesse Ruderman's Quitter *
 * extension to terminate the browser   *
 * cleanly.                             *
 ****************************************/
function quit()
{
  var evt = document.createEvent('Events');
  evt.initEvent('please-quit', true, false);
  document.dispatchEvent(evt);
}
